package org.tekkotsu.ui.editor;

/**
 * @author asangpet
 */

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.gef.ContextMenuProvider;
import org.eclipse.gef.GraphicalViewer;
import org.eclipse.gef.palette.CreationToolEntry;
import org.eclipse.gef.palette.MarqueeToolEntry;
import org.eclipse.gef.palette.PaletteDrawer;
import org.eclipse.gef.palette.PaletteGroup;
import org.eclipse.gef.palette.PaletteRoot;
import org.eclipse.gef.palette.SelectionToolEntry;
import org.eclipse.gef.palette.ToolEntry;
import org.eclipse.gef.requests.SimpleFactory;
import org.eclipse.gef.ui.actions.ActionRegistry;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Sash;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IEditorDescriptor;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorRegistry;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IPathEditorInput;
import org.eclipse.ui.IPersistableElement;
import org.eclipse.ui.IPropertyListener;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.tekkotsu.ui.editor.actions.GenerateLayoutAction;
import org.tekkotsu.ui.editor.actions.CreateNodeAction;
import org.tekkotsu.ui.editor.actions.CreateTransitionAction;
import org.tekkotsu.ui.editor.editparts.MyEditPartFactory;
import org.tekkotsu.ui.editor.model.ModelData;
import org.tekkotsu.ui.editor.model.SourceNodeModel;
import org.tekkotsu.ui.editor.model.SourceTransitionModel;
import org.tekkotsu.ui.editor.model.StateNodeModel;
import org.tekkotsu.ui.editor.model.LayoutData;
import org.tekkotsu.ui.editor.netcode.RobotConnector;
import org.tekkotsu.ui.editor.resources.IDTag;
import org.tekkotsu.ui.rcp.actions.OpenLayoutFileAction;
import org.tekkotsu.ui.util.Debugger;
import org.tekkotsu.ui.util.FilesystemUtil;
import org.tekkotsu.ui.util.RCPUtil;

/**
 * @author asangpet
 */
public class StateMachineEditor extends SinglePageGraphicalEditorWithPalette
		implements PropertyChangeListener {
	public PropertyChangeSupport props = new PropertyChangeSupport(this);
	public static final String DIRTY = "_dirty";
	
	public static String startupLayoutFile = null;
	
	static final int PAGE_PREVIEW = 0;
	static final int PAGE_EDITOR = 1;
	static final int PAGE_SOURCE_MODEL = 2;
	static final int PAGE_SOURCE_LAYOUT = 3;
	
	private Text modelPreviewText;
	private Text behaviorNameText;
	private Text hostText;

	public LayoutData     layoutData;
	public RobotConnector robotConnector = new RobotConnector();

	private PropertyChangeEvent event;
	private boolean dirty_flag = false;

	public static IEditorPart createNewEditor() {
		LayoutData         layoutData = LayoutData.newLayout();
		IPath              path       = layoutData.getPath();
		IEditorPart        editor     = null;
		
		// open editor
		try {
			editor =
				RCPUtil.getPage().openEditor(
					createEditorInput(path),
					getEditorId(path)
				);
		} catch (PartInitException e) {
			Debugger.printThrowable(e);
		}
		
		return editor;
	}

	public void setDirty() {
		dirty_flag = true;
		super.firePropertyChange(PROP_DIRTY);
		props.firePropertyChange(DIRTY, null, null);
	}
	
	public void propertyChange(PropertyChangeEvent evt)
	{
		//the event's text should be appended to the network log
		if (evt.getPropertyName().equals(RobotConnector.EVENT_DATA_UPDATE))
		{
			this.event = evt;
			modelPreviewText.getDisplay().syncExec(
				new Runnable()
				{
					@Override
					public void run()
					{
						modelPreviewText.append(event.getNewValue()+"\r\n");
					}
				}
			);
		}
		
		//we need to update the model's source
		else if (evt.getPropertyName().equals(RobotConnector.EVENT_MODEL_UPDATE))
		{
			String buf = (String) evt.getNewValue();
			Debugger.printDebug(Debugger.DEBUG_ALL,"Downloaded Model XML: " + buf);
				
			//if we downloaded an empty model, alert the user
			if( buf.equals("<model></model>") )
			{
				modelPreviewText.getDisplay().syncExec(
					new Runnable()
					{
						@Override
						public void run()
						{
							MessageDialog.openWarning(
								getContainer().getShell(),
								"Warning: Empty Model",
								"The connection was successful but the model downloaded was empty. Are you sure the name of the behavior is correct?"
							);
						}
					}
			    );
				return;
			}
			
			final StateMachineEditor _this = this;
			layoutData.getModelData().setSourceInput(new StringReader(buf));
			modelPreviewText.getDisplay().syncExec(
				new Runnable()
				{
					@Override
					public void run()
					{
						createModelActions(layoutData);
						layoutData.verifySource();
						if (layoutData.getPartChildren().size()==0)
						{						
							if (MessageDialog.openQuestion(getContainer().getShell(),"New Layout","Would you like the state machine layout to be automatically generated?"))
							{
								GenerateLayoutAction action = new GenerateLayoutAction(_this, layoutData.getModelData());
								action.run();							
							}
						}
					}
				}
		    );
		}
	}

	@SuppressWarnings("unchecked")
	private void createModelActions(LayoutData viewModel) {
		ActionRegistry actionRegistry = this.getActionRegistry();
		List<Action> keyList = new ArrayList<Action>();
		
		{
			Iterator<Action> iter = actionRegistry.getActions();
			while (iter.hasNext()) {
				keyList.add((iter.next()));
			}
			iter = keyList.iterator();
			while (iter.hasNext()) {
				actionRegistry.removeAction(iter.next());
			}
		}
		// actionRegistry.dispose();
		// actionRegistry = new ActionRegistry();

		super.createActions();

		// generate actions for adding state and transition
		{
			Iterator<SourceNodeModel> iter = viewModel.getModelData().getNodes().iterator();
		
			while (iter.hasNext()) {
				SourceNodeModel model = iter.next();
				Action testAction = new CreateNodeAction(this,model);
				testAction.setId(IDTag.ACTION_add_state + model.getId());
				actionRegistry.registerAction(testAction);
			}
		}
		
		{
			Iterator<SourceTransitionModel> iter = viewModel.getModelData().getTransitions().iterator();
			
			while (iter.hasNext()) {
				SourceTransitionModel model = iter.next();
				Action testAction = new CreateTransitionAction(this,model);
				testAction.setId(IDTag.ACTION_add_transition + model.getId());
				actionRegistry.registerAction(testAction);
			}
		}
		Action layoutAction = new GenerateLayoutAction(this, viewModel.getModelData());
		layoutAction.setId(IDTag.ACTION_add_all);
		actionRegistry.registerAction(layoutAction);
	}

	@Override
	protected void createPages() {		
		// initialize editor first		
		GridData gd;
		GridLayout layout;
		
		final Composite parent = this.getContainer();
		layout = new GridLayout();
		layout.numColumns = 1;
		parent.setLayout(layout);
		
		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.horizontalSpan = 1;
		
		createPreviewPage(parent);
		SashForm splitter2 = new SashForm(parent, SWT.HORIZONTAL );
		splitter2.setLayoutData(gd);
		
		createLayoutEditorPage(splitter2);

		Composite logComposite = new Composite(splitter2, SWT.NORMAL);
		layout = new GridLayout();
		logComposite.setLayout(layout);
		layout.marginHeight = 0;
		layout.marginWidth  = 0;
		layout.numColumns = 1;
		layout.verticalSpacing = 0;
		layout.horizontalSpacing = 0;

		gd = new GridData(SWT.FILL, SWT.TOP, true, false);
		gd.horizontalSpan = 1;
		Label label = new Label(logComposite, SWT.NONE);
		label.setText("Network Log");
		label.setLayoutData(gd);
		
		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.horizontalSpan = 1;
		modelPreviewText = new Text(logComposite, SWT.H_SCROLL
				| SWT.V_SCROLL | SWT.BORDER);
		modelPreviewText.setLayoutData(gd);
		
		splitter2.setWeights(new int[] { 80, 20 });
	}
	
	public void createPreviewPage(Composite parent) {
		final StateMachineEditor _this   = this;
	
		GridData   gd;
		GridLayout layout;
		
		Composite allControls      = new Composite(parent, SWT.NORMAL);
		gd = new GridData(SWT.FILL, SWT.TOP, true, false);
		gd.horizontalSpan = 1;
		allControls.setLayoutData(gd);
		layout = new GridLayout();
		layout.verticalSpacing = 0;
		layout.marginHeight = 0;
		allControls.setLayout(layout);
		
		Composite textComposite    = new Composite(allControls, SWT.NORMAL);
		textComposite.setLayoutData(gd);
		layout = new GridLayout();
		textComposite.setLayout(layout);
		layout.numColumns = 6;
		layout.verticalSpacing = 0;
		layout.marginTop = 0;
		
		Composite controlComposite = new Composite(allControls, SWT.NORMAL);
		controlComposite.setLayoutData(gd);
		layout = new GridLayout();
		controlComposite.setLayout(layout);
		layout.numColumns = 5;
		layout.verticalSpacing = 0;
		layout.marginHeight = 0;
	
		gd = new GridData();
		gd.horizontalSpan = 1;
		Label hostLabel = new Label(textComposite, SWT.NONE);
		hostLabel.setText("Host");
		hostLabel.setLayoutData(gd);
	
		gd = new GridData(SWT.FILL, SWT.FILL, true, false);
		gd.horizontalSpan = 1;
		hostText = new Text(textComposite, SWT.BORDER);
		hostText.setLayoutData(gd);
		hostText.setText(robotConnector.getHostName());
		hostText.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent e) {
				robotConnector.setHostName(((Text) e.getSource()).getText());
			}
		});
	
		gd = new GridData();
		gd.horizontalSpan = 1;
		Label labelBehaviorName = new Label(textComposite, SWT.NONE);
		labelBehaviorName.setText("Behavior Name");
		gd.horizontalAlignment = SWT.RIGHT;
		labelBehaviorName.setLayoutData(gd);
	
		gd = new GridData(GridData.FILL_HORIZONTAL);
		gd.horizontalSpan = 1;
		behaviorNameText = new Text(textComposite, SWT.BORDER);
		behaviorNameText.setLayoutData(gd);
		setBehaviorName(robotConnector.getMachineName());
		behaviorNameText.addFocusListener(new FocusAdapter() {
			@Override
			public void focusGained(FocusEvent e) {
				if( behaviorNameText.getText().equals(RobotConnector.defaultMachineName))
				{
					Color col = PlatformUI.getWorkbench().getDisplay().getSystemColor (SWT.COLOR_LIST_BACKGROUND);
					behaviorNameText.setText("");
					behaviorNameText.setBackground(col);
				}
			}
	
			@Override
			public void focusLost(FocusEvent e) {
				setBehaviorName(((Text) e.getSource()).getText());
			}
		});
	
		final Button downloadModelButton = new Button(controlComposite, SWT.NONE);
		final Button saveButton = new Button(controlComposite, SWT.NONE);
		final Button loadButton = new Button(controlComposite, SWT.NONE);
		final Button clearButton = new Button(controlComposite, SWT.NONE);
		final Button reconnectButton = new Button(controlComposite, SWT.NONE);
		
		final Sash sash = new Sash(controlComposite,SWT.HORIZONTAL);
		gd = new GridData(SWT.FILL, SWT.FILL, true, false);
		gd.horizontalSpan = 4;
		sash.setLayoutData(gd);
		
		gd = new GridData(SWT.FILL, SWT.FILL, true, false);
		gd.horizontalSpan = 1;
		
		downloadModelButton.setText("Download Model");
		downloadModelButton.setLayoutData(gd);
		downloadModelButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				// start up client loop
				robotConnector.execute("spider");
			}
		});
		
		clearButton.setText("Clear Layout");
		clearButton.setLayoutData(gd);
		clearButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				layoutData.clear();
			}
		});
		
		reconnectButton.setText("Reconnect");
		reconnectButton.setLayoutData(gd);
		reconnectButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				robotConnector.reconnect();
			}
		});
		
		loadButton.setLayoutData(gd);
		loadButton.setText("Open Layout");
	
		loadButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				//close the current window and open a new one
				OpenLayoutFileAction act = new OpenLayoutFileAction();
				act.run();
			}
		});
	
		saveButton.setText("Save Layout");
		saveButton.setLayoutData(gd);
		saveButton.setEnabled(isDirty());
		saveButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event)
			{
				_this.doSave(new NullProgressMonitor());
			}
		});
		final IPropertyListener SAVE_BUTTON_DIRTY_LISTENER =
			new IPropertyListener()
			{
				@Override
				public void propertyChanged(Object source, int propId)
				{
					saveButton.setEnabled(_this.isDirty());
				}
			};
		this.addPropertyListener(SAVE_BUTTON_DIRTY_LISTENER);
		saveButton.addDisposeListener(
			new DisposeListener()
			{
				@Override
				public void widgetDisposed(DisposeEvent e) {
					_this.removePropertyListener(SAVE_BUTTON_DIRTY_LISTENER);
				}
			}
		);
	}

	/**
	 * @see org.eclipse.ui.IWorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createLayoutEditorPage(Composite parent) {
		createGraphicalViewer(parent);
		createModelActions(layoutData);
	}

	@Override
	protected void configureGraphicalViewer() {
		super.configureGraphicalViewer();
		GraphicalViewer viewer = getGraphicalViewer();
		viewer.setEditPartFactory(new MyEditPartFactory());
	
		// configure the context menu provider
		ContextMenuProvider cmProvider = new StateMachineEditorContextMenuProvider(
				viewer, getActionRegistry());
		viewer.setContextMenu(cmProvider);
		getSite().registerContextMenu(cmProvider, viewer);
	}

	public String getBehaviorName()
	{
		if( behaviorNameText != null )
		{
			String text = behaviorNameText.getText();
			
			if( text == null )
				return null;
			if( text.length() == 0 )
				return null;
			if( text.equals(RobotConnector.defaultMachineName) )
				return null;
			
			return text;
		}
		
		return null;
	}
	
	public void setBehaviorName(String machineName) {
		/*
		 * Get the canonical machine name 
		 */
		if( machineName == null )
		{
			machineName = "";
		}
		
		if( machineName.isEmpty() )
		{
			machineName = RobotConnector.defaultMachineName;
		}
		
		/*
		 * Update the display
		 */
		if( behaviorNameText != null )
		{
			if( machineName.equals(RobotConnector.defaultMachineName) )
			{
				Color col = PlatformUI.getWorkbench().getDisplay().getSystemColor (SWT.COLOR_RED);
				behaviorNameText.setBackground(col);
			}
			else
			{
				Color col = PlatformUI.getWorkbench().getDisplay().getSystemColor (SWT.COLOR_LIST_BACKGROUND);
				behaviorNameText.setBackground(col);
			}
			
			behaviorNameText.setText(machineName);
		}
		
		/*
		 * Update model information
		 */
		if( !machineName.equals(RobotConnector.defaultMachineName) )
		{
			//only update if name is different
			if( !machineName.equals(robotConnector.getMachineName()))
			{
				if (layoutData != null && layoutData.getModelData() != null) {
					ModelData ms = layoutData.getModelData();

					if (!machineName.equals(ms.getMachineName())) {
						ms.setMachineName(machineName);
						setDirty();
					}
				}
			}
		}
		
		try {
			robotConnector.setMachineName(machineName);
		}
		catch (Exception e) {
			Debugger.printThrowable(e);
		}
	}
	
	@Override
	public void doSave(IProgressMonitor monitor) {		
		if (FilesystemUtil.isTempFile(layoutData.getPath())) {
			doSaveAs(getBehaviorName());
		} else {
			doSaveOutput();
		}
	}
	
	private void setSaved() {
		getEditDomain().getCommandStack().markSaveLocation();
		dirty_flag = false;
		firePropertyChange(PROP_DIRTY);
	}
	
	public void updatePartName()
	{
		super.setPartName(layoutData.getEditorTabName());
	}

	@Override
	public void doSaveAs() {
		doSaveAs(getBehaviorName());
	}

	public void doSaveAs(String guess)
	{
		IPath oldpath = layoutData.getPath();
		
		if( layoutData.saveAs(guess) )
		{		
			//change the title of the editor
			updatePartName();
			setSaved();
			
			FilesystemUtil.deleteIfTempFile(oldpath);
		}
	}

	private void doSaveOutput() {
		if( layoutData.save() )
		{
			setSaved();
		}
	}

	public static IEditorInput createEditorInput(IPath path) {
		return new PathEditorInput(path);
	}

	public static String getEditorId(IPath path) {		
		IWorkbench workbench = PlatformUI.getWorkbench();
		IEditorRegistry editorRegistry = workbench.getEditorRegistry();
		IEditorDescriptor descriptor = editorRegistry.getDefaultEditor(path.toFile().getName());
		if (descriptor != null)
			return descriptor.getId();
		
		return "org.tekkotsu.ui.rcp.editors.SimpleEditor"; //$NON-NLS-1$
	}
	
	/**
	 * Generate the layout model from XML document
	 * 
	 * @param document
	 *            view layout file
	 * @return generated layout model for GraphicalViewer
	 */
//	private LayoutData generateLayoutData(IEditorSite site, Document document, IPath path) {
//		Element root = document.getRootElement();	
//
//		LayoutData answer;
//		
//		if ("view".equals(root.getName())) {
//			answer = new LayoutData(site, root, path);
//		}
//		else {
//			answer = new LayoutData(path);
//		}
//		
//		setNewMachineName(answer.getModelData().getMachineName());
//		
//		return answer;
//	}

	@Override
	protected PaletteRoot getPaletteRoot() {
		PaletteRoot root = new PaletteRoot();
		PaletteGroup toolGroup = new PaletteGroup("Tool");

		// Add SelectionTool entry
		ToolEntry tool = new SelectionToolEntry();
		toolGroup.add(tool);
		root.setDefaultEntry(tool);

		// Add MarqueeTool entry
		tool = new MarqueeToolEntry();
		toolGroup.add(tool);

		PaletteDrawer drawer = new PaletteDrawer("Object");
		ImageDescriptor descriptor = ImageDescriptor.createFromFile(
				StateMachineEditor.class, "icons/newModel.gif");
		CreationToolEntry creationEntry = new CreationToolEntry("State node",
				"Create a new state", new SimpleFactory(StateNodeModel.class),
				descriptor, descriptor);
		drawer.add(creationEntry);

		/*
		 * descriptor = ImageDescriptor.createFromFile(StateMachineEditor.class,
		 * "icons/arrowConnection.gif"); CreationToolEntry
		 * arrowConnxCreationEntry = new
		 * ConnectionCreationToolEntry("Transition","Create a new
		 * transition",new
		 * SimpleFactory(TransitionModel.class),descriptor,descriptor);
		 * drawer.add(arrowConnxCreationEntry);
		 */

		root.add(toolGroup);
		root.add(drawer);
		return root;
	}

	@Override
	public void init(IEditorSite site, IEditorInput input) {
		super.init(site, input);
		
		layoutData = LayoutData.loadLayout(new Path(input.getName()));
		updatePartName();
		setBehaviorName(layoutData.getModelData().getMachineName());
		getEditDomain().getCommandStack().markSaveLocation();
	}

	@Override
	protected void initializeGraphicalViewer() {
		GraphicalViewer viewer = getGraphicalViewer();
		viewer.setContents(layoutData);
	}
	
	@Override
	public void dispose()
	{
		FilesystemUtil.deleteIfTempFile(layoutData.getPath());
		robotConnector.disconnect();
		super.dispose();
	}
	
	@Override
	public boolean isDirty() {
		return dirty_flag;
	}

	@Override
	public boolean isSaveAsAllowed() {
		return true;
	}

	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		if (part.getSite().getWorkbenchWindow().getActivePage() == null)
			return;

		super.selectionChanged(part, selection);
	}

	/**
	 * @return Returns the viewModel.
	 */
	public LayoutData getLayoutData() {
		return layoutData;
	}

	/**
	 * @return Returns the modelConnector.
	 */
	public RobotConnector getRobotConnector() {
		return robotConnector;
	}
}

class PathEditorInput implements IPathEditorInput {
	private IPath fPath;
	
	/**
	 * Creates an editor input based of the given file resource.
	 *
	 * @param path the file
	 */
	public PathEditorInput(IPath path) {
		if (path == null) {
			throw new IllegalArgumentException();
		}
		this.fPath = path;
	}
	
	/*
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return fPath.hashCode();
	}
	
	/*
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!(obj instanceof PathEditorInput))
			return false;
		PathEditorInput other = (PathEditorInput) obj;
		return fPath.equals(other.fPath);
	}
	
	/*
	 * @see org.eclipse.ui.IEditorInput#exists()
	 */
	public boolean exists() {
		return fPath.toFile().exists();
	}
	
	/*
	 * @see org.eclipse.ui.IEditorInput#getImageDescriptor()
	 */
	public ImageDescriptor getImageDescriptor() {
		return PlatformUI.getWorkbench().getEditorRegistry().getImageDescriptor(fPath.toString());
	}
	
	/*
	 * @see org.eclipse.ui.IEditorInput#getName()
	 */
	public String getName() {
		return fPath.toString();
	}
	
	/*
	 * @see org.eclipse.ui.IEditorInput#getToolTipText()
	 */
	public String getToolTipText() {
		return fPath.makeRelative().toString();
	}
	
	/*
	 * @see org.eclipse.ui.IPathEditorInput#getPath()
	 */
	public IPath getPath() {
		return fPath;
	}

	/*
	 * @see org.eclipse.core.runtime.IAdaptable#getAdapter(java.lang.Class)
	 */
	@SuppressWarnings("unchecked")
	public Object getAdapter(Class adapter) {
		return null;
	}

	/*
	 * @see org.eclipse.ui.IEditorInput#getPersistable()
	 */
	public IPersistableElement getPersistable() {
		// no persistence
		return null;
	}
}

