/*
 * Created on Nov 11, 2004
 */
package org.tekkotsu.ui.storyboard.panels;

import java.util.ArrayList;
import java.util.Collection;
import java.util.NavigableMap;
import java.util.NavigableSet;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.Graphics;
import org.eclipse.draw2d.Panel;
import org.eclipse.draw2d.geometry.Rectangle;
import org.tekkotsu.ui.rcp.StoryboardView;
import org.tekkotsu.ui.storyboard.model.AbstractModel;
import org.tekkotsu.ui.storyboard.objects.AbstractViewObject;
import org.tekkotsu.ui.storyboard.views.RuntimeView;
import org.tekkotsu.ui.storyboard.views.StoryboardViewer;
import org.tekkotsu.ui.util.Debugger;

/**
 * @author asangpet
 * 
 */

public class Storyboard extends AbstractTimePanel {
	public static final int X_POS = 0;
	public static final int Y_POS = TimeLine.HEIGHT + LogBar.HEIGHT;

	@Override
	protected int getPreferredX() {
		return X_POS;
	}

	@Override
	protected int getPreferredY() {
		return Y_POS;
	}

	/** List of currently selected objects */
	private Set<AbstractViewObject> selectedObjects = new TreeSet<AbstractViewObject>(
			new AbstractViewObject.BeginComparator());

	/** List of all children sorted by their start time */
	private NavigableMap<AbstractViewObject,AbstractViewObject> allStartedObjectsBegin = new TreeMap<AbstractViewObject,AbstractViewObject>(
			new AbstractViewObject.BeginComparator());

	/** Collection of all unfinished children */
	private Collection<AbstractViewObject> allUnfinishedObjects = new TreeSet<AbstractViewObject>(
			new AbstractViewObject.BeginComparator());

	/** List of all finished children sorted by their end time */
	private NavigableSet<AbstractViewObject> allFinishedObjectsEnd = new TreeSet<AbstractViewObject>(
			new AbstractViewObject.EndComparator());

	public Storyboard(Panel parent, StoryboardViewer viewer, TimeMarker marker) {
		super(parent, viewer, marker);
	}

	@Override
	public Rectangle setConstraint(int x, int y, int w, int h) {
		//ensure always tall enough to fill the window
		h = Math.max(h,viewer.getVisibleViewport().height-y);
		
		return super.setConstraint(x, y, w, h);
	}
	
	@Override
	public void timeCursorUpdated() {
		final int CUR_TIME = getTimeCursor();
		RuntimeView.TreeParent infoTree = RuntimeView.getInstance().getContent(CUR_TIME);

		// Extend unfinished objects
		for( AbstractViewObject aobj : allUnfinishedObjects ) {
			aobj.extendToTime(CUR_TIME);
		}
		
		// Save old selection
		Set<AbstractViewObject> oldSelection = new TreeSet<AbstractViewObject>(new AbstractViewObject.BeginComparator());
		oldSelection.addAll(selectedObjects);
		
		// Find selected objects
		selectedObjects.clear();
		findSelection(selectedObjects, CUR_TIME);

		// select objects that weren't selected last update
		for (AbstractViewObject aobj : selectedObjects) {
			if( !oldSelection.contains(aobj) ) {
				aobj.setActive(true);
			}
			infoTree.addChild(aobj.getRuntimeViewData());
		}
		// deselect objects that were selected last update
		for (AbstractViewObject aobj : oldSelection) {
			if( !selectedObjects.contains(aobj) ) {
				aobj.setActive(false);
			}
		}
		StoryboardView.view.runtime.setContent(infoTree);

		super.timeCursorUpdated();
	}

	private void findSelection(Set<AbstractViewObject> selection, int time) {
		try {
			final int CUR_X = getOffsetOfTime(time);
			final int SELECT_BUFFER = 5;
			
			final int SELECTION_BEGIN_TIME = getTimeOfOffset(CUR_X - SELECT_BUFFER);
			final int SELECTION_END_TIME   = getTimeOfOffset(CUR_X + SELECT_BUFFER);			
			
			AbstractViewObject selectionBegin = AbstractViewObject
				.getComparatorQuery(SELECTION_BEGIN_TIME);
			AbstractViewObject selectionEnd = AbstractViewObject
				.getComparatorQuery(SELECTION_END_TIME);

			Collection<AbstractViewObject> beginsBeforeX = allStartedObjectsBegin
					.headMap(selectionEnd, true).values();
			Collection<AbstractViewObject> endsAfterX = allFinishedObjectsEnd
					.tailSet(selectionBegin, true);

			// Find the intersection
			Collection<AbstractViewObject> smaller;
			Collection<AbstractViewObject> larger;

			if (beginsBeforeX.size() > endsAfterX.size()) {
				larger = beginsBeforeX;
				smaller = endsAfterX;
			} else {
				larger = endsAfterX;
				smaller = beginsBeforeX;
			}

			for (AbstractViewObject child : smaller) {
				if (larger.contains(child)) {
					selection.add(child);
				}
			}

			// Manually check unfinished blocks
			Collection<AbstractViewObject> finished = new ArrayList<AbstractViewObject>();
			for (AbstractViewObject child : allUnfinishedObjects) {
				AbstractModel am = child.getModel();
				int left;
				int right;

				left = am.getStart();

				if (am.isFinish()) {
					finished.add(child);
					right = am.getEnd();
				} else {
					right = Integer.MAX_VALUE;
				}

				if (left <= time && right >= time) {
					selection.add(child);
				}
			}
			allUnfinishedObjects.removeAll(finished);
			allFinishedObjectsEnd.addAll(finished);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	protected void paintFigure(Graphics g) {
		Debugger.printDebug("[Storyboard] paintFigure(" + g + ")");
		super.paintFigure(g);

		final Rectangle bounds = getBounds();
		
		// draw cursor
		g.setForegroundColor(ColorConstants.red);

		final int X = getOffsetOfTime(getTimeCursor());
		final int Y1 = bounds.y;
		final int Y2 = bounds.y + bounds.height;

		g.drawLine(X, Y1, X, Y2);
	}
	
	@Override
	public void reset() {
		super.reset();
		super.removeAll();
		selectedObjects.clear();
		allStartedObjectsBegin.clear();
		allUnfinishedObjects.clear();
		allFinishedObjectsEnd.clear();
	}

	@Override
	public void update(AbstractViewObject child) {
		final AbstractViewObject OLD = allStartedObjectsBegin.get(child);
		if( OLD != null ) {
			super.remove(OLD);
			if( OLD.getModel().isFinish() )
			{
				allFinishedObjectsEnd.remove(OLD);
			}
			selectedObjects.remove(OLD);
			allStartedObjectsBegin.remove(OLD);
			allUnfinishedObjects.remove(OLD);
		}

		super.add(child);
		
		Debugger.printDebug("CHILDREN: " + getChildren().size());
		
		final Rectangle CHILD = child.getConstraint();
		final Rectangle CURRENT = getConstraint();

		// position the child
		getLayoutManager().setConstraint(child, CHILD);

		final int CX = CURRENT.x;
		final int CY = CURRENT.y;
		final int CW = CURRENT.width;
		final int CH = CURRENT.height;

		final int R = CHILD.x + CHILD.width;
		final int B = CHILD.y + CHILD.height;

		// make sure storyboard is large enough
		if (R > CW || B > CH) {
			setConstraint(CX, CY, Math.max(R, CW), Math.max(B, CH));
		}

		allStartedObjectsBegin.put(child,child);

		if (child.getModel().isFinish()) {
			allFinishedObjectsEnd.add(child);
		} else {
			allUnfinishedObjects.add(child);
		}
		
		repaint();
	}
}